#ifndef _MCDINTERACTIONS_H /* -*- mode: C; -*- */
#define _MCDINTERACTIONS_H
/*
   Copyright (c) 1997-2002 MathEngine PLC

   $Name:  $

   Date: $Date: 2003/05/05 00:57:07 $ - Revision: $Revision: 1.1 $

   This software and its accompanying manuals have been developed
   by MathEngine PLC ("MathEngine") and the copyright and all other
   intellectual property rights in them belong to MathEngine. All
   rights conferred by law (including rights under international
   copyright conventions) are reserved to MathEngine. This software
   may also incorporate information which is confidential to
   MathEngine.

   Save to the extent permitted by law, or as otherwise expressly
   permitted by MathEngine, this software and the manuals must not
   be copied (in whole or in part), re-arranged, altered or adapted
   in any way without the prior written consent of the Company. In
   addition, the information contained in the software may not be
   disseminated without the prior written consent of MathEngine.

 */

/** @file
 *
 * Queries on the geometrical relationship between a pair of McdModel
 * objects. The principal query is McdIntersect(), which computes contacts.
 *
 * All pairwise queries are performed using McdModelPair objects.  The
 * interactions rely on a retained-mode mechanism: Before a McdModelPair
 * object is used in any of the interaction functions, you must prepare it
 * by calling McdHello(). After that point, the object can be used on any
 * available interaction functions, over multiple time steps in the
 * simulation.
 *
 * The continuity of McdModelPair identity, and keeping track of
 * hello/goodbye events are managed automatically when you obtain your
 * McdModelPair objects using McdSpace.
 *
 * The actual underlying algorithms depend on the actual geometry types of
 * each McdModel object in the pair.
 * The retained-mode mechanism  is exploited by
 * some of these algorithms, which may employ
 * coherence-based techniques, or perform pairwise pre-processing
 * operations when a given pair is first encountered.

 * Both the geometry types and the
 * interaction algorithms must be explicitly registered with the Mcd
 * system at initialization. If, for a particular geometry type - geometry
 * type combination, there is no algorithm registered or available, the
 * interaction will simply be ignored.
 *
 */

#include <McdCTypes.h>
#include <MeInline.h>
#include <MeMath.h>

#ifdef __cplusplus
extern "C" {
#endif


/**
 * Determine quickly if two objects are potentially colliding.
 *
 * Returns non zero value if the object's axis aligned bounding
 * boxes intersect.
 */
MEPUBLIC
MeBool             MEAPI McdNearby( McdModelPair* );

/**
 * Prepare @a p for use by pairwise queries.
 *
 * McdHello() must be called on a pair before using it in any queries,
 * such as McdIntersect() or McdSafeTime().
 *
 * This call should be made on any "hello" pairs obtained from McdSpace
 * for which pairwise queries are to be made.
 *
 * This call may involve allocation internal cached data, or
 * preprocessing some information for use by McdIntersect() or
 * McdSafeTime(), or selection of the appropriate algorithm based on
 * some of the pair's McdRequest values. All data associated with these
 * operations is stored (by reference) in private fields in the
 * McdModelPair object.
 *
 * The order of the two McdModel objects, as obtained from
 * McdModelPairGetModels(), may be inverted after calling McdHello, in
 * order to match the order in which the underlying algorithm has been
 * written. This saves performing an internal "swap" operation each time
 * a pairwise query is invoked. The values in the McdRequest object
 * associated with @p may be examined in order to select an algorithmic
 * specialization or perform some additional preprocessing.
 *
 * After McdHello has been called, the following calls are not allowed
 * on @p until McdGoodbye is called. Otherwise, undefined behaviour
 * results: McdModelPairReset() McdModelPairSetRequestPtr()
 *
 * @see McdIntersect McdSafeTime McdSpaceGetPairs,
 * McdSpaceGetTransitions
 */
MEPUBLIC
void               MEAPI McdHello( McdModelPair* p);

/**
 * Computes the intersection of two McdModel objects, representing it as
 * a set of McdContact objects stored in @a result.
 *
 * The number of contacts computed depends on which geometry type -
 * geometry type combination the pair has, and on the algorithm.
 *
 * The user must allocate @a result and the @a contacts array of the
 * McdIntersectResult struct.  The maximum number of contacts produced
 * is guaranteed to be less than or equal to the number set in the @a
 * contactMaxCount field of McdIntersectResult.
 *
 * The underlying algorithm may also examine the @a request field of
 * McdModelPair, in particular the @a contactMaxCount field of
 * McdRequest. This value represents a user's desired number of contacts
 * to be produced, which may be less than the actual size of the
 * McdContact array.  The algorithm will attempt to meet this request,
 * but it is not guaranteed.
 *
 * This McdRequest value is particularly important for intersections
 * with non-primitive geometry types such as McdTriangleMesh and
 * McdConvexMesh, which may potentially involve a large number of
 * contact points.
 *
 * McdHello() needs to be called on @a p before using this function.
 * The McdRequest object associated with @a p must not change after
 * McdHello() has been called.
 *
 * The McdIntersectResult struct represents an intersection in a format
 * that is appropriate for many types of response, in particular
 * physically-based response. Contact data is used by MathEngine's
 * Dynamics Toolkit, and this function is called directly by
 * MathEngine's Simulation Toolkit.
 *
 * @see McdHello McdIntersectResult
 */
MEPUBLIC
MeBool             MEAPI McdIntersect( McdModelPair*, McdIntersectResult* );

/** 
 * Performs appropriate collision test on input pair at an instant of
 * time in the future specified by @time, extrapolating the current
 * positions of both models using the linear and angular velocities.
 *
 * All data returned in McdIntersectResult "result."
 *
 * The return value is 1 if the proper collision function was available,
 * 0 otherwise.
 *
 * McdHello needs to be called on this pair before calling this
 * function.
 *
 * @see McdIntersectResult McdSpaceGetStayingPairs
 */
MEPUBLIC
MeBool             MEAPI McdIntersectAt( McdModelPair *p,
                        McdIntersectResult *result, MeReal time );

/**
 * Performs appropriate SafeTime (time of impact and swept volume)
 * computation on input pair.
 *
 * All data is returned in McdSafeTimeResult "result".  Models are
 * assumed to be synchronized, that is, their transforms correspond to
 * positions at identical instants of time.  The model's linear and
 * angular velocities are used to describe the motion of the object.
 *
 * The input value of maxTime indicates the maximum time for travel
 * given the model linear and angular velocities.
 *
 * The timestep for a dynamical simulation is a typical value for
 * maxTime.  For a value of 1, the object is assumed to translate by the
 * entire linear velocity vector.
 *
 * The return value is 1 if the proper SafeTime function was available,
 * 0 otherwise.
 *
 * This function must be called after McdHello(p)
 * This function is subject to change.
 *
 * @see McdSafeTimeResult for a description of the data returned.
 */
MEPUBLIC
MeBool             MEAPI McdSafeTime( McdModelPair*,
                        MeReal maxTime, McdSafeTimeResult* );

/**
 * Call McdIntersect on all "hello" and "staying" pairs in @a pairs.
 *
 * Must call McdHello on each of these pairs first, or simply call
 * McdHelloEach() with @a pairs as argument.
 *
 * User allocates an array of McdIntersectResult structs and McdContact
 * structs for the results to be written into. ( @a resultArray, @a
 * resultArraySize, @a contactArray, @a contactArraySize ).
 *
 * As each pair is processed, it writes to as many of the next unwritten
 * elements in @a contactArray as needed, and writes to the next
 * unwritten element in @a resultArray. McdIntersectResult::contacts
 * field will point to the location of the first element in @a
 * contactArray that was used.
 *
 * The total number of elements in @a resultArray that were written to,
 * and the total number of element in @a contactArray that were written
 * to are set in the output arguments @a resultCount and @a
 * contactCount, respectively.
 *
 * The return value indicates overflow:
 *
 * If either @a resultArray or @a contactArray were not large enough to
 * allow all the appropriate pairs in @a pairs to be processed, a value
 * of 0 is returned, and @a pairsIter keeps track of which pairs in @a
 * pairs have not been processed. In this case, process the results
 * computed, and then call the function again until no overflow remains.
 *
 * This batch mode mechanism allows you to choose the workspace memory
 * size you can afford to use, and then takes care of the bookkeeping to
 * use that space effectively. A larger workspace means a larger number
 * of pairs can be processed in "the same batch".
 */
MEPUBLIC
MeBool             MEAPI McdIntersectEach( McdModelPairContainer* pairs,
                       McdModelPairContainerIterator* pairsIter,
                       McdIntersectResult* resultArray,int resultArraySize,
                       int *resultCount,
                       McdContact* contactArray,int contactArraySize,
                       int *contactCount);
/** To be documented */
MeINLINE void
McdContactReverse(McdContact *cp)
{
    int e1 = cp->element1.tag;
    cp->element1.tag = cp->element2.tag;
    cp->element2.tag = e1;
    cp->dims = (cp->dims>>8) + ((cp->dims&0xff)<<8);

    MeVector3Scale(cp->normal,-1);
}

/** To be documented */
MeINLINE void
McdIntersectResultReverse(McdIntersectResult *result)
{
    int i;
    MeVector3Scale(result->normal,-1);

    for(i=0;i<result->contactCount;i++)
        McdContactReverse(result->contacts+i);
}

/**
 * Intersects an oriented line segment with a single model.
 *
 * @arg cm the collision model.
 * @arg inOrig pointer to an MeVector3 representing the first point on
 *      the line segment.
 * @arg inDest pointer to an MeVector3 representing the second point on
 *      the line segment.
 * @arg outOverlap structure containing the line segment intersection
 *      data.
 *
 * @see McdSpaceGetLineSegmentOverlapList
 */
MEPUBLIC
unsigned int        MEAPI McdLineSegIntersect(const McdModelID cm,
                        MeReal* const inOrig, MeReal* const inDest,
                        McdLineSegIntersectResult *outOverlap);

#ifdef __cplusplus
}
#endif


#endif
